Option Explicit

'
' Database pruning script. Usage:
'
'
' cscript prunedb.vbs 
'
'   - Reports how many players / ip addresses would be deleted 
'     for not playing for 30 days
'
' cscript prunedb.vbs /delete
'
'   - Actually deletes players / ip addresses that
'     have not played for 30 days
'
' cscript prunedb.vbs 60
'
'   - Reports how many players / ip addresses would be deleted 
'     for not playing for 60 days
'
' cscript prunedb.vbs 60 /delete
'
'   - Actually deletes players / ip addresses that
'     have not played for 60 days
'
'
' NOTE: Players that have notes attached or have played more than 1000
'       rounds are never deleted by this script.


Dim fdir
Dim fname
Dim fso 
Dim xmlIn
Dim ipdict
Dim netdict

Dim gKillCount
Dim gIPKillCount
Dim gKeepCount
Dim gIPKeepCount


Dim gCutoff
gCutoff = 30

gKillCount = 0
gIPKillCount = 0

Set fso = CreateObject("Scripting.FileSystemObject")
Set xmlIn = CreateObject("MSXML2.DomDocument.3.0")
Set ipdict = CreateObject("Scripting.Dictionary")
Set netdict = CreateObject("Scripting.Dictionary")

xmlIn.Async = False

Dim fDelete
fDelete = False

if wscript.arguments.count > 0 Then
  If wscript.arguments.count = 2 Then
    If wscript.arguments(1) = "/delete" Then
      fDelete = True
    End If
    If Clng("0" & wscript.arguments(0)) > 0 Then
      gCutoff = Clng("0" & wscript.arguments(0))
    End If
  Else
    if wscript.arguments(0) = "/delete" Then
      fDelete = True
    Else
      If Clng("0" & wscript.arguments(0)) > 0 Then
        gCutoff = Clng("0" & wscript.arguments(0))
      End If
    end if
  End If
end if


For Each fdir In fso.GetFolder("db").SubFolders

  wscript.echo "Scaning players: " & fdir.Name 
  For Each fname In fdir.Files
  	If right(fname.Name,15)="Connections.xml" Then  		
  		If DateDiff("d",fname.DateLastModified,Now) > gCutoff Then
  			ParseFile fname
  		Else
  			ParseIPs fname
  			gKeepCount = gKeepCount + 1
  		End If  	
  	End If
  Next
Next



For Each fdir In fso.GetFolder("ips").SubFolders
  wscript.echo "Scaning addresses: " & fdir.Name 
  For Each fname In fdir.Files
  	If right(fname.Name,11)="_subnet.txt" Then  		
		If Not netdict.Exists(Left(fname.Name,Len(fName.name)-11)) Then
			DeleteFile fName.Path
			gIPKillCount = gIPKillCount + 1
		Else
			gIPKeepCount = gIPKeepCount + 1
		End If
	ElseIf right(fname.Name,12)="_address.txt" Then  		
		If Not ipdict.Exists(Left(fname.Name,Len(fName.name)-12)) Then
			DeleteFile fName.Path
			gIPKillCount = gIPKillCount + 1
		Else
			gIPKeepCount = gIPKeepCount + 1
		End If
	End If
  Next
Next

If fDelete Then
	wscript.echo "Deleted " & gKillCount & " players, kept " & gKeepCount
	wscript.echo "Deleted " & gIPKillCount & " IPs, kept " & gIPKeepCount
Else
	wscript.echo "Would remove " & gKillCount & " players and keep " & gKeepCount
	wscript.echo "Would remove " & gIPKillCount & " IPs and keep " & gIPKeepCount
End If

Sub ParseFile(f)
  Dim n
  Dim count
  
  xmlIn.Load f.Path
  If xmlIn.parseError.errorCode = 0 Then
    count=0	
    For Each n In xmlIn.SelectNodes("connections/connection/count")
      count = count + CLng("0" & n.text)
    Next
    
    'Anyone who has played 1000 lives forever!
    If count > 1000 Then
      ParseIPs f
      gKeepCount = gKeepCount + 1
      Exit Sub
    End If
  End If
  
  xmlIn.Load Replace(f.Path,"_Connections.xml","_Notes.xml")
  If xmlIn.parseError.errorCode = 0 Then
    For Each n In xmlIn.SelectNodes("notes/note")
      'Anyone who has a note lives forever
      ParseIPs f
      gKeepCount = gKeepCount + 1
      Exit Sub
    Next
  End If
  
  gKillCount = gKillCount + 1
  
  Dim strPath
  strPath = fName.path
  Set fName = Nothing
  
  DeleteFile strPath
  DeleteFile Replace(strPath,"_Connections.xml","_IPs.xml")
  DeleteFile Replace(strPath,"_Connections.xml","_Kills.xml")
  DeleteFile Replace(strPath,"_Connections.xml","_Namess.xml")
  DeleteFile Replace(strPath,"_Connections.xml","_Notes.xml")
  DeleteFile Replace(strPath,"_Connections.xml",".xml")
    
End Sub

Sub ParseIPs(f)
  Dim n
  Dim nPos
  Dim strIP
  
  xmlIn.Load Replace(f.Path,"_Connections.xml","_IPs.xml")
  If xmlIn.parseError.errorCode = 0 Then
    For Each n In xmlIn.SelectNodes("ips/ip/address")
      strIP = n.Text  
      ipdict(strIP)=1
      npos = instr(strIP,".")
      npos = instr(npos+1,strIP,".")
      npos = instr(npos+1,strIP,".")
      if npos = 0 then
      	wscript.echo "wtf"
      Else
      	netdict(left(strIP,nPos-1) & ".0") = 1
      End If
    Next
  End If
End Sub

Sub DeleteFile(strPath)
  If fDelete Then
    If fso.FileExists(strPath) Then
      fso.DeleteFile strPath
    End If
  End If
End Sub